
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Random;

class Sample {

    public Short leftSample;
    public Short rightSample;

    public Sample(Short left, Short right) {
        leftSample = left;
        rightSample = right;
    }

}

class WaveData {

    private ArrayList<Sample> audioSamples;

    public WaveData() {
        audioSamples = new ArrayList<>();
    }

    public WaveData(String filePath) {
        audioSamples = new ArrayList<>();
        File file = new File(filePath);
        if (!file.isFile())
            return;
        parseFileBytes(file);
    }

    public WaveData(File file) {
        audioSamples = new ArrayList<>();
        if (!file.isFile())
            return;
        parseFileBytes(file);
    }

    private void parseFileBytes(File fileToParse) {

        // lettura del fileToParse byte per byte
        InputStream input = null;
        try {
            input = new FileInputStream(fileToParse);
        } catch (FileNotFoundException ex) {
            return;
        }

        int dataChunkSize = (int) (fileToParse.length() - 44);
        byte[] data = new byte[dataChunkSize];
        try {
            input.skip(44); // salto la lettura dei primi 44 byte
            for (int i = 0; i < dataChunkSize; i++)
                data[i] = (byte) input.read();
        } catch (IOException ex) {
            return;
        }

        // conversione dell'array di byte in array di short
        int bufferSize = (int) (dataChunkSize / 2);
        short[] result = new short[bufferSize];
        ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN).asShortBuffer().get(result);

        for (int i = 0; i < result.length; i = i + 2) {
            Sample s = new Sample(result[i], result[i + 1]);
            audioSamples.add(s);
        }
    }

    private byte[] returnHeader() {

        byte[] header = new byte[44];

        header[0] = 'R';
        header[1] = 'I';
        header[2] = 'F';
        header[3] = 'F';

        int audioSamplesByteSize = audioSamples.size() * 4;
        int chunkSize = (int) (36 + audioSamplesByteSize);
        // bit shifting e maschera di bit
        header[4] = (byte) (chunkSize & 0xff);
        header[5] = (byte) ((chunkSize >> 8) & 0xff);
        header[6] = (byte) ((chunkSize >> 16) & 0xff);
        header[7] = (byte) ((chunkSize >> 24) & 0xff);

        header[8] = 'W';
        header[9] = 'A';
        header[10] = 'V';
        header[11] = 'E';

        // Sub-chunk fmt ----------------------------------------------------
        // header
        header[12] = 'f';
        header[13] = 'm';
        header[14] = 't';
        header[15] = ' ';

        // subchunk size
        header[16] = 16; // little endian
        header[17] = 0;
        header[18] = 0;
        header[19] = 0;

        // Audio format: PCM
        header[20] = 1; // little endian
        header[21] = 0;

        // Number of channels
        int numChannels = 2;
        header[22] = (byte) (numChannels & 0xff); // little endian
        header[23] = (byte) ((numChannels >> 8) & 0xff);

        // Sample rate
        int sampleRate = 44100;
        header[24] = (byte) (sampleRate & 0xff);
        header[25] = (byte) ((sampleRate >> 8) & 0xff);
        header[26] = (byte) ((sampleRate >> 16) & 0xff);
        header[27] = (byte) ((sampleRate >> 24) & 0xff);

        // Byte rate
        int bitsPerSample = 16;
        int byteRate = sampleRate * numChannels * bitsPerSample / 8;
        header[28] = (byte) (byteRate & 0xff);
        header[29] = (byte) ((byteRate >> 8) & 0xff);
        header[30] = (byte) ((byteRate >> 16) & 0xff);
        header[31] = (byte) ((byteRate >> 24) & 0xff);

        // Block align
        int blockAlign = numChannels * bitsPerSample / 8;
        header[32] = (byte) (blockAlign & 0xff);
        header[33] = (byte) ((blockAlign >> 8) & 0xff);

        // Bit depth
        header[34] = (byte) (bitsPerSample & 0xff);
        header[35] = (byte) ((bitsPerSample >> 8) & 0xff);

        // Sub-chunk data ----------------------------------------------------
        // header
        header[36] = 'd';
        header[37] = 'a';
        header[38] = 't';
        header[39] = 'a';

        // subchunk size
        header[40] = (byte) (audioSamplesByteSize & 0xff);
        header[41] = (byte) ((audioSamplesByteSize >> 8) & 0xff);
        header[42] = (byte) ((audioSamplesByteSize >> 16) & 0xff);
        header[43] = (byte) ((audioSamplesByteSize >> 24) & 0xff);

        return header;
    }

    public boolean saveSamplesToDisk(String pathFileName) {

        File file = new File(pathFileName);
        if (!file.exists())
            try {
                file.createNewFile();
            } catch (IOException ex) {
                return false;
            }

        byte[] bytes = new byte[audioSamples.size() * 4];
        for (int i = 0; i < audioSamples.size(); i++) {
            bytes[4 * i] = (byte) (audioSamples.get(i).leftSample & 0xff);
            bytes[4 * i + 1] = (byte) ((audioSamples.get(i).leftSample >> 8) & 0xff);
            bytes[4 * i + 2] = (byte) (audioSamples.get(i).rightSample & 0xff);
            bytes[4 * i + 3] = (byte) ((audioSamples.get(i).rightSample >> 8) & 0xff);
        }

        // scrittura su disco tramite FileWriter e BufferedWriter
        try {
            FileOutputStream fos = new FileOutputStream(file);
            fos.write(returnHeader());
            fos.write(bytes);
            fos.close();
        } catch (IOException ex) {
            return false;
        }
        return true;
    }

    public void generateWhiteNoise(int nSamples, double normalizedAmplitude, boolean clear) {
        if (clear)
            audioSamples.clear();

        if (Math.abs(normalizedAmplitude) > 1)
            normalizedAmplitude = 1;

        Random rand = new Random();
        for (int i = 0; i < nSamples; i++) {
            short leftCandidate = (short) ((rand.nextInt(65535) - 32768) * Math.abs(normalizedAmplitude));
            short rightCandidate = (short) ((rand.nextInt(65535) - 32768) * Math.abs(normalizedAmplitude));
            Sample s = new Sample(leftCandidate, rightCandidate);
            audioSamples.add(s);
        }
    }

    public void generateSquare(int nSamples, double normalizedAmplitude, double frequency, boolean clear) {
        if (clear)
            audioSamples.clear();

        if (Math.abs(normalizedAmplitude) > 1)
            normalizedAmplitude = 1;
        short amplitude = (short) (32767 * Math.abs(normalizedAmplitude));

        double nSamplesPerCycle = 44100 / frequency;
        for (int k = 0; k < nSamples; k++) {
            Sample s;
            if ((k % nSamplesPerCycle) < (nSamplesPerCycle / 2))
                s = new Sample(amplitude, amplitude);
            else
                s = new Sample((short) -amplitude, (short) -amplitude);
            audioSamples.add(s);
        }
    }

    public void generateSawTooth(int nSamples, double normalizedAmplitude, double frequency, boolean clear) {
        if (clear)
            audioSamples.clear();

        if (Math.abs(normalizedAmplitude) > 1)
            normalizedAmplitude = 1;
        short amplitude = (short) (32767 * Math.abs(normalizedAmplitude));

        double nSamplesPerCycle = 44100 / frequency;
        for (int k = 0; k < nSamples; k++) {
            int currentSample = (int) (k % nSamplesPerCycle);
            short candidate = (short) ((currentSample * (amplitude * 2) / nSamplesPerCycle) - amplitude);
            Sample s = new Sample(candidate, candidate);
            audioSamples.add(s);
        }
    }

    public void generateSine(int nSamples, double normalizedAmplitude, double frequency, boolean clear) {
        if (clear)
            audioSamples.clear();

        if (Math.abs(normalizedAmplitude) > 1)
            normalizedAmplitude = 1;
        short amplitude = (short) (32767 * Math.abs(normalizedAmplitude));

        double nSamplesPerCycle = 44100 / frequency;
        for (int k = 0; k < nSamples; k++) {
            short candidate = (short) (Math.sin(2 * Math.PI * k / nSamplesPerCycle) * amplitude); // (k / nSamplesPerCycle) calcola il punto corrente sulla circonferenza goniometrica per il calcolo del seno
            Sample s = new Sample(candidate, candidate);
            audioSamples.add(s);
        }
    }

    public void clip(double normalizedAmplitude) {
        if (Math.abs(normalizedAmplitude) > 1)
            normalizedAmplitude = 1;

        short clipAmplitude = (short) (32767 * Math.abs(normalizedAmplitude));
        for (Sample s : audioSamples) {
            if (s.leftSample > clipAmplitude)
                s.leftSample = clipAmplitude;
            if (s.leftSample < -clipAmplitude)
                s.leftSample = (short) -clipAmplitude;
            if (s.rightSample > clipAmplitude)
                s.rightSample = clipAmplitude;
            if (s.rightSample < -clipAmplitude)
                s.rightSample = (short) -clipAmplitude;
        }
    }

    public void normalize(double normalizedAmplitude) {
        if (Math.abs(normalizedAmplitude) > 1)
            normalizedAmplitude = 1;

        short maxAmplitude = 0;
        for (Sample s : audioSamples) {
            if (Math.abs(s.leftSample) > maxAmplitude)
                maxAmplitude = (short) Math.abs(s.leftSample);
            if (Math.abs(s.rightSample) > maxAmplitude)
                maxAmplitude = (short) Math.abs(s.rightSample);
        }

        double rescale = Math.abs(normalizedAmplitude) * 32767.0 / maxAmplitude;
        for (Sample s : audioSamples) {
            s.leftSample = (short) (s.leftSample * rescale);
            s.rightSample = (short) (s.rightSample * rescale);
        }
    }

    public void normalize() {
        normalize(1);
    }

    public void setLevels(double normalizedLeftLevel, double normalizedRightLevel) {
        if (Math.abs(normalizedLeftLevel) > 1)
            normalizedLeftLevel = 1;
        if (Math.abs(normalizedRightLevel) > 1)
            normalizedRightLevel = 1;

        for (Sample s : audioSamples) {
            s.leftSample = (short) (s.leftSample * Math.abs(normalizedLeftLevel));
            s.rightSample = (short) (s.rightSample * Math.abs(normalizedRightLevel));
        }
    }
}

public class WavUtils1 {

    public static void main(String[] args) throws InterruptedException, IOException {
        WaveData wd = new WaveData("pini.wav");
        wd.normalize();
        wd.saveSamplesToDisk("pini2.wav");
    }
}
